from django.conf import settings
from django.http import HttpResponse
from django.views.decorators.cache import never_cache
from blti import BLTI, BLTIException
from blti.performance import log_response_time
import json
import re


class RESTDispatchAuthorization(Exception):
    pass


class RESTDispatchMethod(Exception):
    pass


class RESTDispatch(object):
    """ A superclass for views, that handles passing on the request to the
        appropriate view method, based on the request method.
    """

    extra_response_headers = {}

    @never_cache
    @log_response_time
    def run(self, *args, **named_args):
        try:
            request = args[0]
            self.authorize(request)
            return self.dispatch(request.method)(*args, **named_args)
        except RESTDispatchAuthorization as ex:
            return self.error_response(401, ex)
        except RESTDispatchMethod:
            return self.error_response(405, 'Method not allowed')

    def _http_response(self, content, *args, **kwargs):
        response = HttpResponse(content, *args, **kwargs)
        for k, v in self.extra_response_headers.iteritems():
            response[k] = v

        return response

    def get_session(self, request):
        return BLTI().get_session(request)

    def authorize(self, request):
        self.blti_authorize(request)

    def blti_authorize(self, request):
        try:
            self.get_session(request)
        except BLTIException as ex:
            if not (getattr(settings, 'BLTI_NO_AUTH', False) and
                    request.user.is_authenticated()):
                raise RESTDispatchAuthorization('%s' % ex)

    def dispatch(self, method):
        methods = dict((m, m) for m in ['GET', 'HEAD', 'POST', 'PUT',
                                        'DELETE', 'PATCH', 'OPTIONS'])
        try:
            return getattr(self, methods[method])
        except (KeyError, AttributeError):
            raise RESTDispatchMethod()

    def error_response(self, status, message='', content={}):
        content['error'] = str(message)
        return self.json_response(content=content, status=status)

    def json_response(self, content={}, status=200):
        return self._http_response(json.dumps(content),
                                   status=status,
                                   content_type='application/json')



import time
from logging import getLogger
from blti import BLTI


logger = getLogger(__name__)


def log_response_time(func):
    def wrapper(*args, **kwargs):
        self = args[0]

        start = time.time()
        try:
            val = func(*args, **kwargs)
        except:
            raise
        finally:
            module = self.__module__
            function = func.__name__

            is_view_class = True
            if module == "django.core.handlers.wsgi":
                is_view_class = False

            if is_view_class:
                arg_str = str(args[2:])
            else:
                arg_str = str(args[1:])

            kw_str = str(kwargs)
            end = time.time()

            try:
                request = args[1]
                login_id = BLTI().get_session(request).get(
                    'custom_canvas_user_login_id')
            except Exception as ex:
                login_id = 'unknown_user'

            arg_str = arg_str.replace("#", "___")
            kw_str = kw_str.replace("#", "___")

            margs = (login_id,
                     module,
                     function,
                     arg_str,
                     kw_str,
                     end-start)

            msg = "user# %s method# %s.%s args# %s kwargs# %s time# %s" % margs
            logger.info(msg)
        return val
    return wrapper



from django.db import models


class BLTIKeyStore(models.Model):
    consumer_key = models.CharField(max_length=80, unique=True)
    shared_secret = models.CharField(max_length=80)
    added_date = models.DateTimeField(auto_now_add=True)



from django.conf import settings
from blti.models import BLTIKeyStore
from blti import BLTIException
import oauth2 as oauth
import re


class BLTIOauth(object):
    def validate(self, request, params={}):
        oauth_server = oauth.Server()
        oauth_server.add_signature_method(
            oauth.SignatureMethod_HMAC_SHA1())

        oauth_request = oauth.Request.from_request(
            request.method,
            request.build_absolute_uri(),
            headers=request.META,
            parameters=params
        )

        if oauth_request:
            try:
                key = oauth_request.get_parameter('oauth_consumer_key')
                consumer = self.get_consumer(key)
                oauth_server._check_signature(oauth_request, consumer, None)
                return oauth_request.get_nonoauth_parameters()
            except oauth.Error as err:
                raise BLTIException(err)

        raise BLTIException('Invalid OAuth Request')

    def get_consumer(self, key):
        try:
            model = BLTIKeyStore.objects.get(consumer_key=key)
            return oauth.Consumer(key, str(model.shared_secret))

        except BLTIKeyStore.DoesNotExist:
            try:
                consumers = getattr(settings, 'LTI_CONSUMERS', {})
                return oauth.Consumer(key, consumers[key])
            except KeyError:
                return None


class BLTIRoles(object):
    ADMIN = 'admin'
    MEMBER = 'member'
    ALL = None

    # https://www.imsglobal.org/specs/ltiv1p1/implementation-guide#toc-19
    LIS_ADMIN = [
        'AccountAdmin', 'SysAdmin', 'SysSupport', 'Faculty', 'Staff',
        'Creator', 'Administrator',
        'Administrator/Administrator', 'Administrator/Developer',
        'Administrator/ExternalDeveloper', 'Administrator/ExternalSupport',
        'Administrator/ExternalSystemAdministrator', 'Administrator/Support',
        'Administrator/SystemAdministrator',
        'Manager', 'Manager/AreaManager', 'Manager/CourseCoordinator',
        'Manager/ExternalObserver', 'Manager/Observer'
    ]

    LIS_INSTRUCTOR = [
        'Instructor',
        'Instructor/ExternalInstructor', 'Instructor/GuestInstructor',
        'Instructor/Lecturer', 'Instructor/PrimaryInstructor',
        'TeachingAssistant',
        'TeachingAssistant/Grader', 'TeachingAssistant/TeachingAssistant',
        'TeachingAssistant/TeachingAssistantGroup',
        'TeachingAssistant/TeachingAssistantOffering',
        'TeachingAssistant/TeachingAssistantSection',
        'TeachingAssistant/TeachingAssistantSectionAssociation',
        'TeachingAssistant/TeachingAssistantTemplate'
    ]

    LIS_LEARNER = [
        'Alumni', 'Guest', 'Learner', 'Member', 'ProspectiveStudent',
        'Student', 'Learner', 'Learner/ExternalLearner',
        'Learner/GuestLearner', 'Learner/Instructor', 'Learner/Learner',
        'Learner/NonCreditLearner', 'Member', 'Member/Member'
    ]

    def _has_role(self, roles, lis_roles):
        for role in roles:
            if role in lis_roles:
                return True

            m = re.match(r'^urn:lti:(inst|sys)?role:ims/lis/([A-Za-z]+)$',
                         role)
            if m and m.group(2) in lis_roles:
                return True

        return False

    def has_admin_role(self, roles):
        return self._has_role(roles, self.LIS_ADMIN)

    def has_instructor_role(self, roles):
        return self._has_role(roles, self.LIS_INSTRUCTOR)

    def has_learner_role(self, roles):
        return self._has_role(roles, self.LIS_LEARNER)

    def validate(self, blti, visibility):
        if visibility:
            roles = ','.join([blti.get('roles', ''),
                              blti.get('ext_roles', '')]).split(',')

            # ADMIN includes instructors, MEMBER includes
            if not (self.has_admin_role(roles) or
                    self.has_instructor_role(roles) or
                    (visibility == self.MEMBER and
                    self.has_learner_role(roles))):
                raise BLTIException(
                    'You do not have privilege to view this content.')



"""
Cross Site Request Forgery and Session HTTP Header Middleware.

This module provides a middleware that implements protection
against request forgeries from other sites.
"""

from django.conf import settings


class CSRFHeaderMiddleware(object):
    def process_view(self, request, callback, callback_args, callback_kwargs):
        csrf_token = request.META.get('HTTP_X_CSRFTOKEN', None)
        if csrf_token is not None:
            csrf_token_name = settings.CSRF_COOKIE_NAME
            request.COOKIES[csrf_token_name] = csrf_token

        return None


class SessionHeaderMiddleware(object):
    def process_request(self, request):
        session_id = request.META.get('HTTP_X_SESSIONID', None)
        if session_id is not None:
            session_key = settings.SESSION_COOKIE_NAME
            request.COOKIES[session_key] = session_id



from django.views.generic.base import TemplateView
from django.views.decorators.csrf import csrf_exempt
from blti import BLTI, BLTIException
from blti.validators import BLTIOauth, BLTIRoles
try:
    from urllib import unquote_plus
except ImportError:
    from urllib.parse import unquote_plus  # Python3


class BLTIView(TemplateView):
    http_method_names = ['get', 'options']

    def get(self, request, *args, **kwargs):
        try:
            params = self.validate(request)
        except BLTIException as err:
            self.template_name = 'blti/401.html'
            return self.render_to_response({})

        context = self.get_context_data(
            request=request, blti_params=params, **kwargs)

        response = self.render_to_response(context)
        self.add_headers(response=response, blti_params=params, **kwargs)
        return response

    def get_context_data(self, **kwargs):
        return kwargs

    def add_headers(self, **kwargs):
        pass

    def get_session(self, request):
        return BLTI().get_session(request)

    def set_session(self, request, **kwargs):
        BLTI().set_session(request, **kwargs)

    def validate(self, request):
        return self.get_session(request)


class BLTILaunchView(BLTIView):
    http_method_names = ['post']
    authorized_role = 'member'

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        return super(BLTILaunchView, self).dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        try:
            params = self.validate(request)
            self.set_session(request, **params)
        except BLTIException as err:
            self.template_name = 'blti/error.html'
            return self.render_to_response({'error': err})

        context = self.get_context_data(
            request=request, blti_params=params, **kwargs)

        response = self.render_to_response(context)
        self.add_headers(response=response, blti_params=params, **kwargs)
        return response

    def validate(self, request):
        params = {}
        body = request.read()
        if body and len(body):
            params = dict((k, v) for k, v in [tuple(
                map(unquote_plus, kv.split('='))
            ) for kv in body.split('&')])
        else:
            raise BLTIException('Missing or malformed parameter or value')

        blti_params = BLTIOauth().validate(request, params=params)
        if blti_params:
            BLTIRoles().validate(blti_params, visibility=self.authorized_role)

        return blti_params


class RawBLTIView(BLTILaunchView):
    template_name = 'blti/raw.html'
    authorized_role = 'admin'
